-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module severs as API for the logic used in the toolkit

module Assertion.LogicAPI where

import Assertion.PresburgerValidity
import Common.Absglobaltype
import Common.Utils
import Data.Set
import Data.Map as M
import Data.List as List
import Common.PrettyPrinter
--
-- Returns True if an expression is satisfiable
--
checkSatis :: Sorting -> Exp -> Bool
checkSatis _ BTrue = True
checkSatis s exp = checkPresburger (mkUniversal s exp)



--
-- checkSatis wrapper for Assertions
--
checkAssertion :: Sorting -> Assertion -> Bool
checkAssertion s (Assert exp) = checkSatis s exp


--
-- Universally quantifies an expression
--
mkUniversal :: Sorting -> Exp -> Exp
mkUniversal sort e = case varlist of
                       [] -> e
                       xs -> EForall (Prelude.map (\(x, y) -> (VarDeclaration x y)) xs) e                                        
    where varlist = M.toList $ variableMap sort e

--
-- Existentially quantifies an expression
-- 
mkExistential :: Sorting -> Exp -> Exp
mkExistential sort e = case varlist of
                         [] -> e
                         xs -> EExist (Prelude.map (\(x, y) -> (VarDeclaration x y)) xs) e
    where varlist = M.toList $ variableMap sort e


-- MOVED to utils.hs
-- --
-- -- Free variable (and their types) from an expression
-- --
-- variableMap :: Sorting -> Exp -> Map Variable Typ
-- variableMap sort (ExpV v) = case (lookupVariable sort v) of 
--                               Just x -> M.singleton v x
--                               Nothing -> error $ "Variable "++(show v)++" not found."
-- variableMap sort (BNot e) = variableMap sort e
-- variableMap sort (EBin e1 b e2) = M.union (variableMap sort e1) (variableMap sort e2)
-- variableMap sort (EAr e1 b e2) = M.union (variableMap sort e1) (variableMap sort e2)
-- variableMap sort (EExist v e) = (variableMap (addVariables sort v) e) 
--                                 `M.difference` (M.fromList (List.map (\(VarDeclaration vr t) -> (vr,t)) v))
-- variableMap sort (EForall v e) = (variableMap (addVariables sort v) e) 
--                                  `M.difference` (M.fromList (List.map (\(VarDeclaration vr t) -> (vr,t)) v)) 
-- variableMap sort e = M.empty