-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module is used as a LIBRARY
-- by the programs generated by Venocid
-- it used to synch the concurrent children
-- of a program
--
module CodeGeneration.ProcManager where

import Control.Concurrent
import Control.Exception


--
-- Block the main thread until all the children terminates
-- (see 
-- http://www.haskell.org/ghc/docs/latest/html/libraries/base/Control-Concurrent.html
-- for more information)
--
waitForChildren :: MVar [MVar ()] -> IO ()
waitForChildren children = do
  cs <- takeMVar children
  case cs of
    []   -> return ()
    m:ms -> do
            putMVar children ms
            takeMVar m
            waitForChildren children


--
-- Create a new child and add it in the "global variable"
--
forkChild :: MVar [MVar ()] -> IO () -> IO ThreadId
forkChild children io = do
    mvar <- newEmptyMVar
    childs <- takeMVar children
    putMVar children (mvar:childs)
    forkIO (io `finally` putMVar mvar ())
