-- Author: Julien Lange (jl250@le.ac.uk)
-- Partially generated by the BNF converter
-- 2009-2010

-- This modules contains the Abstract Syntax
-- for all the languages used by Venocid
-- (i.e. Global, local types and processes)
-- See language documentation for more information
--
-- The module also contains general types
-- used (notably) for typing
--

module Common.Absglobaltype where

import Data.Map as M

{- ------------------------ -}
{-     LANGUAGE TYPES       -}
{- ------------------------ -}

newtype Ident = Ident String deriving (Eq,Ord)

instance Show Ident where
    show (Ident i) = i

data Program =
   GlobalProg [GlobalSpecification]
 | SingleProg Process
 | Prog [GlobalSpecification] Process
  deriving (Eq,Ord,Show)

data GlobalSpecification =
   Global Channel DistributedInteraction
  deriving (Eq,Ord,Show)

data DistributedInteraction =
   SingleInter Interaction
 | ParallelInter [ConcurInteraction]
  deriving (Eq,Ord,Show)

data ConcurInteraction =
   Conc DistributedInteraction
  deriving (Eq,Ord,Show)

data Interaction =
   ValPass Participant Participant Channel [VarDecl] Assertion DistributedInteraction
 | Deleg Participant Participant Channel DelegDecl Assertion DistributedInteraction
 | LabelPass Participant Participant Channel BranchId [BranchElmt]
 | End
 | Recur Variable [Exp] [VarDeclLoc] Assertion DistributedInteraction
 | RecCall Variable [Exp]
  deriving (Eq,Ord,Show)

data BranchElmt =
   BranchingElmt Assertion Label DistributedInteraction
  deriving (Eq,Ord,Show)

data Process =
   SingleProcess Primitive
 | ParallelProcess [ConcurProcess]
  deriving (Eq,Ord,Show)

data ConcurProcess =
   ConcP Process
  deriving (Eq,Ord,Show)

data Primitive =
   PInit Channel [Participant] [Channel] Process
 | PJoin Channel Participant [Channel] Process
 | PSend Channel [Exp] [VarDecl] Assertion Process
 | PDelegS Channel [Channel] DelegDecl Assertion Process
 | PReceive Channel [VarDecl] Assertion Process
 | PDelegR Channel DelegDecl Assertion Process
 | PSelect Channel Assertion BranchId Label Process
 | PBranch Channel BranchId [PBranchElmt]
 | PEnd
 | ITE Exp Process Process
 | Hiding Channel Process
 | PRecur Variable [Exp] [Channel] [VarDecl] [Channel] Process
 | PRecCall Variable [Exp] [Channel]
  deriving (Eq,Ord,Show)

data PBranchElmt =
   PBranchElmt Assertion Label Process
  deriving (Eq,Ord,Show)

data Assertion =
   Assert Exp
 | AssertEmpty
  deriving (Eq,Ord,Show)

data Exp =
   ExpI Integer
 | ExpV Variable
 | ExpS String
 | ExpD Integer Integer Integer
 | BTrue
 | BFalse
 | BNot Exp
 | EBin Exp BinOp Exp
 | EAr Exp ArOp Exp
 | EExist [VarDecl] Exp
 | EForall [VarDecl] Exp
  deriving (Eq,Ord,Show)


data BinOp =
   EOr
 | EAnd
 | EEq
 | ENEq
 | EGt
 | ELt
 | ELte
 | EGte
 | EImpl
  deriving (Eq,Ord,Show)

data ArOp =
    EPlus
        | EMinus
        | EMul
        | EDiv
  deriving (Eq,Ord,Show)


data LocalType =
   Send Channel [VarDecl] Assertion LocalType
 | LDelegS Channel DelegDecl Assertion LocalType
 | Receive Channel [VarDecl] Assertion LocalType
 | LDelegR Channel DelegDecl Assertion LocalType
 | Branch Channel BranchId [LabelElmt]
 | Select Channel BranchId [LabelElmt]
 | LRecur Variable [Exp] [VarDecl] Assertion LocalType
 | LRecCall Variable [Exp]
 | LEnd
 | Bottom
  deriving (Eq,Ord,Show)

data LabelElmt =
   LabelElmt Assertion Label LocalType
  deriving (Eq,Ord,Show)

data DelegDecl =
   SessPass [Channel] LocalType Participant
  deriving (Eq,Ord,Show)

data Typ =
   TInt
 | TBool
 | TString
 | TDate
  deriving (Eq,Ord,Show)

data Participant =
   Parti Ident
  deriving (Eq,Ord,Show)

data Channel =
   Chan Ident
  deriving (Eq,Ord,Show)



data Label =
   Lab Ident
  deriving (Eq,Ord,Show)

data BranchId =
   BId Ident
  deriving (Eq,Ord,Show)

data Variable =
   Var Ident
  deriving (Eq,Ord,Show)

data VarDecl =
   VarDeclaration Variable Typ
  deriving (Eq,Ord,Show)

data VarDeclLoc =
   VarDeclLoc VarDecl Location
  deriving (Eq,Ord,Show)

data Location =
   Loc Participant Participant
  deriving (Eq,Ord,Show)




{- ------------------------ -}
{-    PROJECTION TYPES      -}
{- ------------------------ -}

-- (public_name, [shared_names],projected_type)
type Typing = [(Channel, Participant, [Channel], LocalType)]

type RecurTyping = ([VarDecl], [Channel], Process)

--
-- Sorting
--
data Sorting =  Empty | Sorting NameMap ProcessMap ProjMap
                deriving (Eq,Ord,Show)

--
-- Environment to keep track of
-- * Projection
-- * Processes (in recursion)
-- * Variable
-- in Local typing
--
--
newtype NameMap = NameMap (Map Variable Typ)
    deriving (Eq,Ord,Show)

emptyNameMap :: NameMap
emptyNameMap = NameMap empty

newtype ProcessMap = ProcessMap (Map Variable RecurTyping)
    deriving (Eq,Ord,Show)

emptyProcessMap :: ProcessMap
emptyProcessMap = ProcessMap empty

newtype ProjMap = ProjMap (Map Channel [Projection])
    deriving (Eq,Ord,Show)

emptyProjMap :: ProjMap
emptyProjMap = ProjMap empty

data Projection = Proj Participant LocalType
                  deriving (Eq,Ord,Show)



--
-- Return the type of a variable
--
lookupVariable :: Sorting -> Variable -> Maybe Typ
lookupVariable Empty v = Nothing
lookupVariable (Sorting (NameMap nm) _ _) v = M.lookup v nm
-- Nothing ->  error ((show v)++" is not declared in "++(show $ M.keys nm)++".")
-- Just x -> x

--
-- Return the type of a process
-- (used for recursive processes)
--
lookupProcess :: Sorting -> Variable -> RecurTyping
lookupProcess Empty v =  error ("no type for"++(show v)++".")
lookupProcess (Sorting _ (ProcessMap pm) _) v = case M.lookup v pm of
                                                Nothing ->  error ((show v)++" is not declared in "++(show $ M.keys pm)++".")
                                                Just x -> x

--
-- Add a recursive process type to an environment
--
addProcess :: Sorting -> Variable -> RecurTyping -> Sorting
addProcess Empty v t = Sorting emptyNameMap (ProcessMap $ singleton v t) emptyProjMap
addProcess (Sorting nm (ProcessMap procm) projm) v t = Sorting nm (ProcessMap $ insert v t procm) projm


--
-- Add a variable to an environment
--
addVariable :: Sorting -> Variable -> Typ -> Sorting
addVariable Empty v t = Sorting (NameMap $ singleton v t) emptyProcessMap emptyProjMap
addVariable (Sorting (NameMap s) pm prm )  v t = 
    if member v s
    then error ("Variable "++(show v)++" is declared more than once in "++(show s))
    else Sorting (NameMap (insert v t s)) pm prm


--
-- Add a variables to an environment
--

addVariables :: Sorting -> [VarDecl] -> Sorting
addVariables s [] = s
addVariables s ((VarDeclaration v t):xs) = addVariables (addVariable s v t) xs


--
-- Return the local type corresponding to a tupe (channel,participant)
-- (i.e. a projection for (shared name, participant) )
--
lookupLocalType :: Sorting -> Channel -> Participant -> Maybe LocalType
lookupLocalType Empty _ _ = Nothing
lookupLocalType (Sorting _ _ (ProjMap pm)) k p = case M.lookup k pm of
                                                 Nothing -> Nothing
                                                 Just x -> findProjection x p


--
-- Add a projection map to an environment
--
addProjectionMap :: ProjMap -> Sorting
addProjectionMap p =  Sorting (NameMap empty) (ProcessMap empty) p



{- ------------------------------------------------------------------- -}
{-                           SYMBOL TABLE                              -}
{- ------------------------------------------------------------------- -}

--
-- SymbolTable used in the parser
-- TODO: should be used insted of sort
--
type SymbolTable = Map Variable Typ

newst :: SymbolTable
newst = M.empty

stlookup :: SymbolTable -> Variable -> Maybe Typ

stlookup t v = M.lookup v t


stsingle :: Variable -> Typ -> SymbolTable
stsingle v t = M.singleton v t

stinsert :: SymbolTable -> Variable -> Typ -> SymbolTable
stinsert t v ty = case stlookup t v of
                    Nothing -> M.insert v ty t
                    Just x -> error $ "Variable "++(show v)++" is declared more than once."


stinserts :: SymbolTable -> [(Variable, Typ)] -> SymbolTable
stinserts st [] = st
stinserts st ((v, t):xs) = stinsert (stinserts st xs) v t

--
-- Returns the local type (maybe) of a participant
--
findProjection :: [Projection] -> Participant -> Maybe LocalType
findProjection [] p = Nothing -- error ("Projection for "++(show p)++" not found.")
findProjection ((Proj p lt):xs) q = if p == q
                                    then Just lt
                                    else findProjection xs q



