-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module provides functions to type expressions



module Common.ExpressionTyping
    (expTyping, expsTyping) 
    where

import Common.Absglobaltype
-- import Common.Typing
import Common.PrettyPrinter


--
-- Type an expression (Exp) accordint to its environment (Sort)
--
expTyping :: Sorting -> Exp -> Maybe Typ
expTyping _ (ExpI e) = Just TInt
expTyping s (ExpV e) = lookupVariable s e
expTyping _ (ExpS e) = Just TString
expTyping _ (ExpD _ _ _) = Just TDate
expTyping _ (BTrue) = Just TBool
expTyping _ (BFalse) = Just TBool
expTyping s (BNot e) = if (expTyping s e) == Just TBool
                       then Just TBool
                       else Nothing
expTyping s e@(EBin e1 EOr e2) =
    if (expTyping s e1) == Just TBool && (expTyping s e2) == Just TBool
    then Just TBool
    else Nothing
expTyping s e@(EBin e1 EAnd e2) =
    if (expTyping s e1) == Just TBool && (expTyping s e2) == Just TBool
    then Just TBool
    else Nothing
expTyping s e@(EBin e1 EEq e2) =
    if (expTyping s e1) == (expTyping s e2) && not ((expTyping s e1) == Nothing)
    then Just TBool
    else Nothing
expTyping s e@(EBin e1 ENEq e2) =
    if (expTyping s e1) == (expTyping s e2) && not ((expTyping s e1) == Nothing)
    then Just TBool
    else Nothing
expTyping s e@(EAr e1 EPlus e2) =
    if (expTyping s e1) == (expTyping s e2) && not ((expTyping s e1) == Nothing)
    then Just TInt
    else Nothing
expTyping s e@(EAr e1 EMinus e2) =
    if (expTyping s e1) == (expTyping s e2) && not ((expTyping s e1) == Nothing)
    then Just TInt
    else Nothing
expTyping s e@(EAr e1 EDiv e2) =
    if (expTyping s e1) == (expTyping s e2) && not ((expTyping s e1) == Nothing)
    then Just TInt
    else Nothing
expTyping s e@(EAr e1 EMul e2) =
    if (expTyping s e1) == (expTyping s e2) && not ((expTyping s e1) == Nothing)
    then Just TInt
    else Nothing
expTyping s (EBin e1 b e2) = expTyping_helper s e1 e2 (prPrint b)
expTyping s (EExist vd e) = 
    if (expTyping (addVariables s vd) e) == Just TBool
    then Just TBool
    else Nothing
expTyping s (EForall vd e) = 
    if (expTyping (addVariables s vd) e) == Just TBool
    then Just TBool
    else Nothing
expTyping _ e = error $ "[ExpTyping]Unknown expression:"++show e

-- for EGt ELt ELte EGte
expTyping_helper s e1 e2 txt = 
    if (expTyping s e1) == Just TInt && (expTyping s e2) == Just TInt
    then Just TBool
    else error ("wrong type in "++txt++" "++(prPrint e1)++" and "++(prPrint e2))

--
-- Types a list of expressions
--
expsTyping :: Sorting -> [Exp] -> [Maybe Typ] -- types a list of expression (used in Send)
expsTyping _ [] = []
expsTyping sort (x:xs) = (expTyping sort x):(expsTyping sort xs)