module Main where

import Control.Concurrent
import Control.Concurrent.Chan
import Control.Exception
--
-- Need to find a way to send a list of values: 
--  -might have to use "atomic" and several send/receive
--  -or just send a list [a,b,c]? or tuple ?
--


data MyType = TInt Int | TString String | TBool Bool | TDate String | TLabel String
              deriving (Eq,Ord,Show)

-- Thread termination synch

-- like it is said on
-- http://www.haskell.org/ghc/docs/latest/html/libraries/base/Control-Concurrent.html
-- it lets you block the main thread until all the children terminates
waitForChildren :: MVar [MVar ()] -> IO ()
waitForChildren children = do
  cs <- takeMVar children
  case cs of
    []   -> return ()
    m:ms -> do
            putMVar children ms
            takeMVar m
            waitForChildren children


-- creates a new thread within the thread syncrhonization mechanism
forkChild :: MVar [MVar ()] -> IO () -> IO ThreadId
forkChild children io = do
    mvar <- newEmptyMVar
    childs <- takeMVar children
    putMVar children (mvar:childs)
    forkIO (io `finally` putMVar mvar ())

-- [End] Thread termination synch

main = do
  children <- newMVar [] -- synch
  s1 <- newChan
  b <- newChan
  s2 <- newChan
  n <- newChan
  forkChild children $ do -- Buyer
         putStrLn "Starting Buyer..."
         writeChan n $ TString "War and Peace"
         price <- readChan s2
         writeChan s1 price
         sel <- readChan b
         if (sel == TLabel "enough")
          then do
           writeChan n $ TInt 12
           days <- readChan s2
           putStrLn "Buyer1 Done."
          else do
           writeChan n $ TInt 0
           days <- readChan s2 
           putStrLn "Buyer2 Done."
 
  forkChild children $ do -- Seller
        putStrLn "Starting Seller..."
        title <- readChan n
        writeChan s2 $ TInt 150
        house <- readChan n
        if (house > TInt 15)
         then writeChan s2 $ TInt 5
         else writeChan s2 $ TInt 10
        putStrLn "Seller Done."

  forkChild children $ do -- Bank
         putStrLn "Starting Bank..."
         money <- readChan s1
         if (money > TInt 101)
          then writeChan b $ TLabel "enough"
          else writeChan b $ TLabel "notenough"
         putStrLn "Bank Done."

  waitForChildren children -- synch
  putStrLn "End."